@@ -50,6 +50,7 @@ typedef struct _GripGestureClassMap
 /*
  * GripGestureManagerPrivate:
  * @geis: the GEIS instance
+ * @is_initialized: GEIS async initialization has been successfully completed
  * @iochannel: connects the geis object with the main loop
  * @iochannel_id: the identifier for the iochannel
  * @devices: a collection of recognized input devices
@@ -60,6 +61,7 @@ typedef struct _GripGestureClassMap
 struct _GripGestureManagerPrivate
 {
   Geis              geis;
+  gboolean          is_initialized;
   GIOChannel       *iochannel;
   int               iochannel_id;
   GripDevices      *devices;
@@ -258,6 +260,7 @@ grip_geis_event_callback(Geis geis G_GNUC_UNUSED, GeisEvent geis_event, void *co
       break;
 
     case GEIS_EVENT_INIT_COMPLETE:
+      priv->is_initialized = TRUE;
       break;
 
     case GEIS_EVENT_GESTURE_BEGIN:
@@ -532,8 +535,77 @@ tap_gesture_handle_properties (GripEventGestureTap *event,
 }
 
 
+/**
+ * Gets the extents of a widget in root window coordinates.
+ */
 static void
-process_gesture_events(GripGestureManager *manager, GeisEvent geis_event)
+get_absolute_widget_extents (GtkWidget *widget, gint *left, gint *right, gint *top, gint *bottom)
+{
+  GdkWindow* widget_window = gtk_widget_get_window (widget);
+
+  gdk_window_get_origin (widget_window, left, top);
+
+  gint window_width = gdk_window_get_width (widget_window);
+  gint window_height = gdk_window_get_height (widget_window);
+  gdk_window_get_root_coords (widget_window, window_width, window_height, right, bottom);
+}
+
+
+/**
+ * Gets the focus of a gesture frame in root window coordinates.
+ */
+static void
+get_absolute_frame_focus (GeisFrame frame, gint *x, gint *y)
+{
+  *x = 0;
+  *y = 0;
+
+  GeisAttr focus_x_attr = geis_frame_attr_by_name (frame, GEIS_GESTURE_ATTRIBUTE_FOCUS_X);
+  if (focus_x_attr)
+    {
+      *x = (int)geis_attr_value_to_float(focus_x_attr);
+    }
+
+  GeisAttr focus_y_attr = geis_frame_attr_by_name (frame, GEIS_GESTURE_ATTRIBUTE_FOCUS_Y);
+  if (focus_y_attr)
+    {
+      *y = (int)geis_attr_value_to_float(focus_y_attr);
+    }
+}
+
+
+/**
+ * Indicates if the focus of a gesture frame is inside a widget.
+ */
+static gboolean
+is_frame_focus_in_widget (GeisFrame frame, GtkWidget *widget)
+{
+  gboolean frame_focus_is_in_widget = FALSE;
+
+  GtkWidget *toplevel_widget = gtk_widget_get_toplevel (widget);
+  guint toplevel_widget_window_id = GDK_WINDOW_XID (gtk_widget_get_window (toplevel_widget));
+
+  GeisAttr window_attr = geis_frame_attr_by_name(frame, GEIS_GESTURE_ATTRIBUTE_EVENT_WINDOW_ID);
+  guint window_id = geis_attr_value_to_integer(window_attr);
+
+  if (window_id == toplevel_widget_window_id)
+    {
+      gint widget_left, widget_top, widget_right, widget_bottom;
+      get_absolute_widget_extents (widget, &widget_left, &widget_right, &widget_top, &widget_bottom);
+
+      gint focus_x, focus_y;
+      get_absolute_frame_focus (frame, &focus_x, &focus_y);
+
+      frame_focus_is_in_widget = (focus_x >= widget_left && focus_x <= widget_right
+                               && focus_y >= widget_top  && focus_y <= widget_bottom);
+    }
+
+  return frame_focus_is_in_widget;
+}
+
+
+static void
+process_gesture_events (GripGestureManager *manager, GeisEvent geis_event)
 {
   GripGestureManagerPrivate *priv = manager->priv;
 
@@ -561,9 +633,6 @@ process_gesture_events(GripGestureManager *manager, GeisEvent geis_event)
           GeisFrame frame = geis_group_frame(group, j);
           g_return_if_fail (frame);
 
-          GeisAttr window_attr = geis_frame_attr_by_name(frame, GEIS_GESTURE_ATTRIBUTE_EVENT_WINDOW_ID);
-          guint window_id = geis_attr_value_to_integer(window_attr);
-
           GeisAttr device_attr = geis_frame_attr_by_name(frame, GEIS_GESTURE_ATTRIBUTE_DEVICE_ID);
           guint device_id = geis_attr_value_to_integer(device_attr);
           GripInputDevice *input_device = device_id_to_input_device (manager, device_id);
@@ -572,8 +641,7 @@ process_gesture_events(GripGestureManager *manager, GeisEvent geis_event)
           for (l = priv->bindings; l != NULL; l = l->next)
             {
               GripGestureBinding *binding = (GripGestureBinding *)l->data;
-              guint widget_window_id = GDK_WINDOW_XID (gtk_widget_get_window (binding->widget));
-              if (widget_window_id == window_id)
+              if (is_frame_focus_in_widget(frame, binding->widget))
                 {
                   GList *class_list;
                   for (class_list = priv->classes; class_list != NULL; class_list = class_list->next)
@@ -660,6 +728,7 @@ grip_gesture_manager_init (GripGestureManager *self)
   GripGestureManagerPrivate *priv = self->priv = GRIP_GESTURE_MANAGER_GET_PRIVATE (self);
   priv->devices = g_ptr_array_new ();
 
+  priv->is_initialized = FALSE;
   priv->geis = geis_new (GEIS_INIT_TRACK_DEVICES, GEIS_INIT_TRACK_GESTURE_CLASSES, NULL);
   g_return_if_fail (priv->geis);
 
@@ -722,6 +791,8 @@ processed_mapped_window_request (GripRegistrationRequest *req)
   g_return_if_fail (GTK_IS_WINDOW (toplevel));
 
   GripGestureManagerPrivate *priv = req->manager->priv;
+  if (!priv->is_initialized)
+    return;
 
   guint window_id = GDK_WINDOW_XID (gtk_widget_get_window(toplevel));
 
